// XMSAALib.cpp  Version 1.0 - see article at CodeProject.com
//
// Author:  Hans Dietrich
//          hdietrich@gmail.com
//
// Description:
//     XMSAALib implements a class to retrieve MSAA information from a window.
//
// History
//     Version 1.0 - 2009 February 10
//     - Initial public release
//
// License:
//     This software is released under the Code Project Open License (CPOL),
//     which may be found here:  http://www.codeproject.com/info/eula.aspx
//     You are free to use this software in any way you like, except that you 
//     may not sell this source code.
//
//     This software is provided "as is" with no expressed or implied warranty.
//     I accept no liability for any damage or loss of business that this 
//     software may cause.
//
///////////////////////////////////////////////////////////////////////////////

//#include "stdafx.h"		// use of MFC is optional
#include <windows.h>
#include <tchar.h>
#include <crtdbg.h>
#include "XMSAALib.h"
#include "comutil.h"

#pragma comment(lib, "oleacc.lib")
#pragma comment(lib, "comsupp.lib")

#ifndef __noop
#if _MSC_VER < 1300
#define __noop ((void)0)
#endif
#endif

#undef TRACE
#define TRACE __noop

// if you want to see the TRACE output, 
// uncomment this line:
//#include "XTrace.h"

#pragma warning(disable : 4127)	// for _ASSERTE: conditional expression is constant
#pragma warning(disable : 4996)	// disable bogus deprecation warning

#define MOZILLA_FIREFOX_APP_WINDOW		_T("MozillaUIWindowClass")
#define INTERNET_EXPLORER_APP_WINDOW	_T("IEFrame")		// for IE6 and IE7
#define GOOGLE_CHROME_APP_WINDOW		_T("Chrome_XPFrame")
#define OPERA_APP_WINDOW				_T("OpWindow")

#define SAFE_RELEASE(p_)					{ if (p_ != NULL) p_->Release(); p_ = NULL; }

static XWINEVENTURLPROC g_lpfnXWinEventUrlProc = 0;

//=============================================================================
CXMSAALib::CXMSAALib() :
//=============================================================================
	m_eHook(0)
{
	g_lpfnXWinEventUrlProc = 0;
}

//=============================================================================
CXMSAALib::~CXMSAALib()
//=============================================================================
{
	StopGetObjectFromEvent();
}

//=============================================================================
static void CALLBACK CXMSAALib_WinEventProc(HWINEVENTHOOK /*hook*/, 
											DWORD event, 
											HWND hwnd,
											LONG idObject, 
											LONG idChild,
											DWORD /*dwEventThread*/, 
											DWORD /*dwmsEventTime*/)
//=============================================================================
{
	switch (event)
	{
		case EVENT_OBJECT_FOCUS:
		{
			TRACE(_T("event=0x%X\n"), event);

			IAccessible *pIAcc = 0;
			VARIANT varChild;
			VariantInit(&varChild);

			HRESULT hr = AccessibleObjectFromEvent(hwnd, idObject, idChild, 
							&pIAcc, &varChild);
	
			if ((hr == S_OK) && (pIAcc != NULL)) 
			{
				if (IsWindow(hwnd))
				{
					TRACE(_T("hwnd=0x%X\n"), hwnd);
					TCHAR szClassname[256];
					::GetClassName(hwnd, szClassname, sizeof(szClassname)/sizeof(TCHAR)-1);
					szClassname[sizeof(szClassname)/sizeof(TCHAR)-1] = 0;
					TRACE(_T("class %s\n"), szClassname);
				}

				if (g_lpfnXWinEventUrlProc)
				{
					g_lpfnXWinEventUrlProc(pIAcc, &varChild, event, hwnd);
					TRACE(_T("after event proc\n"));
				}

				SAFE_RELEASE(pIAcc);
			}
			break;
		}

		default:
			break;
	}
}

//=============================================================================
BOOL CXMSAALib::StartGetObjectFromEvent(XWINEVENTURLPROC lpfnXWinEventUrlProc)
//=============================================================================
{
	TRACE(_T("in CXMSAALib::StartGetObjectFromEvent\n"));

	g_lpfnXWinEventUrlProc = 0;

	if (lpfnXWinEventUrlProc)
	{
		g_lpfnXWinEventUrlProc = lpfnXWinEventUrlProc;
	}

	_ASSERTE(g_lpfnXWinEventUrlProc);

	if (g_lpfnXWinEventUrlProc == 0)
	{
		StopGetObjectFromEvent();
	}
	else if (g_lpfnXWinEventUrlProc)
	{
		if (!m_eHook)
		{
			m_eHook = ::SetWinEventHook(EVENT_OBJECT_FOCUS, EVENT_OBJECT_VALUECHANGE, 0, 
				CXMSAALib_WinEventProc, 0, 0, WINEVENT_OUTOFCONTEXT | WINEVENT_SKIPOWNPROCESS);
		}
	}

	TRACE(_T("exiting CXMSAALib::StartGetObjectFromEvent\n"));
	return m_eHook != 0;
}

//=============================================================================
void CXMSAALib::StopGetObjectFromEvent()
//=============================================================================
{
	if (m_eHook)
		::UnhookWinEvent(m_eHook);
	m_eHook = 0;
	g_lpfnXWinEventUrlProc = 0;
}

//=============================================================================
HWND CXMSAALib::GetFirefoxHwnd()
//=============================================================================
{
	return ::FindWindow(MOZILLA_FIREFOX_APP_WINDOW, 0);
}

//=============================================================================
HWND CXMSAALib::GetInternetExplorerHwnd()
//=============================================================================
{
	return ::FindWindow(INTERNET_EXPLORER_APP_WINDOW, 0);
}

//=============================================================================
HWND CXMSAALib::GetGoogleChromeHwnd()
//=============================================================================
{
	return ::FindWindow(GOOGLE_CHROME_APP_WINDOW, 0);
}

//=============================================================================
HWND CXMSAALib::GetOperaHwnd()
//=============================================================================
{
	return ::FindWindow(OPERA_APP_WINDOW, 0);
}

//=============================================================================
BOOL CXMSAALib::GetObjectName(IAccessible *pAcc, 
							  VARIANT *pvarChild, 
							  LPTSTR lpszName, 
							  UINT cchName)
//=============================================================================
{
	BOOL bRet = FALSE;

	_ASSERTE(pAcc);
	_ASSERTE(pvarChild);
	_ASSERTE(lpszName);

	if (pAcc && pvarChild && lpszName)
	{
		lpszName[0] = 0;
		BSTR bstrName = NULL;
		
		HRESULT hr = pAcc->get_accName(*pvarChild, &bstrName);
		
		if (SUCCEEDED(hr) && bstrName)
		{
			_bstr_t btemp(bstrName, false);
			if ((LPCTSTR)btemp)
				_tcsncpy(lpszName, btemp, cchName);
			bRet = TRUE;
		}
	}
	
	return bRet;
} 

//=============================================================================
BOOL CXMSAALib::GetObjectValue(IAccessible *pAcc, 
							   VARIANT *pvarChild, 
							   LPTSTR lpszValue, 
							   UINT cchValue)
//=============================================================================
{
	BOOL bRet = FALSE;

	_ASSERTE(pAcc);
	_ASSERTE(pvarChild);
	_ASSERTE(lpszValue);

	if (pAcc && pvarChild && lpszValue)
	{
		lpszValue[0] = 0;
		BSTR bstrValue = NULL;
		
		HRESULT hr = pAcc->get_accValue(*pvarChild, &bstrValue);
		
		if (SUCCEEDED(hr) && bstrValue)
		{
			_bstr_t btemp(bstrValue, false);
			if ((LPCTSTR)btemp)
				_tcsncpy(lpszValue, btemp, cchValue);
			bRet = TRUE;
		}
	}
	
	return bRet;
} 

//=============================================================================
BOOL CXMSAALib::GetObjectDescription(IAccessible *pAcc, 
									 VARIANT *pvarChild, 
									 LPTSTR lpszDescription, 
									 UINT cchDescription)
//=============================================================================
{
	BOOL bRet = FALSE;

	_ASSERTE(pAcc);
	_ASSERTE(pvarChild);
	_ASSERTE(lpszDescription);

	if (pAcc && pvarChild && lpszDescription)
	{
		lpszDescription[0] = 0;
		BSTR bstrDescription = NULL;
		
		HRESULT hr = pAcc->get_accDescription(*pvarChild, &bstrDescription);
		
		if (SUCCEEDED(hr) && bstrDescription)
		{
			_bstr_t btemp(bstrDescription, false);
			if ((LPCTSTR)btemp)
				_tcsncpy(lpszDescription, btemp, cchDescription);
			bRet = TRUE;
		}
	}
	
	return bRet;
} 

//=============================================================================
BOOL CXMSAALib::GetObjectChildCount(IAccessible *pAcc, 
									long& nCount)
//=============================================================================
{
	BOOL bRet = FALSE;

	_ASSERTE(pAcc);

	if (pAcc)
	{
		HRESULT hr = pAcc->get_accChildCount(&nCount);
		
		if (SUCCEEDED(hr))
		{
			bRet = TRUE;
		}
	}
	
	return bRet;
} 
//=============================================================================
BOOL CXMSAALib::GetObjectLocation(IAccessible *pAcc,
								  VARIANT *pvarChild, 
								  RECT& rect)
//=============================================================================
{
	BOOL bRet = FALSE;

	_ASSERTE(pAcc);
	_ASSERTE(pvarChild);

	if (pAcc && pvarChild)
	{
		HRESULT hr = pAcc->accLocation(&rect.left, &rect.top, 
						&rect.right, &rect.bottom, *pvarChild);
		
		if (SUCCEEDED(hr))
		{
			// accLocation returns width and height
			rect.right += rect.left;
			rect.bottom += rect.top;
			bRet = TRUE;
		}
	}
	
	return bRet;
} 

//=============================================================================
BOOL CXMSAALib::GetObjectRole(IAccessible *pAcc, 
							  VARIANT *pvarChild, 
							  UINT& nRole)
//=============================================================================
{
	BOOL bRet = FALSE;
	
	nRole = 0;
	
	_ASSERTE(pAcc);
	_ASSERTE(pvarChild);

	if (pAcc && pvarChild)
	{
		VARIANT varRole;
		VariantInit(&varRole);
		
		HRESULT hr = pAcc->get_accRole(*pvarChild, &varRole);
		
		if (SUCCEEDED(hr) && (varRole.vt == VT_I4))
		{
			nRole = varRole.lVal;
			bRet = TRUE;
		}
		
		VariantClear(&varRole);
	}
	
	return bRet;
}

//=============================================================================
BOOL CXMSAALib::GetObjectRoleString(IAccessible *pAcc, 
									VARIANT *pvarChild, 
									LPTSTR lpszBuf,
									UINT cchBuf)
//=============================================================================
{
	BOOL bRet = FALSE;

	_ASSERTE(pAcc);
	_ASSERTE(pvarChild);
	_ASSERTE(lpszBuf);

	if (pAcc && pvarChild && lpszBuf)
	{
		lpszBuf[0] = 0;
		UINT nRole = 0;

		VARIANT varRole;
		VariantInit(&varRole);
		
		HRESULT hr = pAcc->get_accRole(*pvarChild, &varRole);
		
		if (SUCCEEDED(hr))
		{
			if (varRole.vt == VT_I4)
			{
				nRole = varRole.lVal;
				::GetRoleText(nRole, lpszBuf, cchBuf);
				bRet = TRUE;
			}
			else if (varRole.vt == VT_BSTR)
			{
				_bstr_t brole(varRole.bstrVal, TRUE);	// copy
				if ((LPCTSTR)brole)
					_tcsncpy(lpszBuf, (LPCTSTR)brole, cchBuf);
				bRet = TRUE;
			}
			VariantClear(&varRole);
		}
	}
	return bRet;
}

//=============================================================================
BOOL CXMSAALib::GetObjectState(IAccessible* pAcc, 
							   VARIANT* pvarChild, 
							   UINT& nState)
//=============================================================================
{
	BOOL bRet = FALSE;

	nState = 0;
	
	_ASSERTE(pAcc);
	_ASSERTE(pvarChild);

	if (pAcc && pvarChild)
	{
		VARIANT varState;
		VariantInit(&varState);
		
		HRESULT hr = pAcc->get_accState(*pvarChild, &varState);
		
		if (SUCCEEDED(hr) && (varState.vt == VT_I4))
		{
			nState = varState.lVal;
			bRet = TRUE;
		}
		
		VariantClear(&varState);
	}
	
	return bRet;
}

//=============================================================================
BOOL CXMSAALib::GetObjectStateString(IAccessible *pAcc, 
									 VARIANT *pvarChild, 
									 LPTSTR lpszBuf,
									 UINT cchBuf)
//=============================================================================
{
	BOOL bRet = FALSE;

	_ASSERTE(pAcc);
	_ASSERTE(pvarChild);
	_ASSERTE(lpszBuf);

	if (pAcc && pvarChild && lpszBuf)
	{
		lpszBuf[0] = 0;
		UINT nState = 0;
		if (GetObjectState(pAcc, pvarChild, nState))
		{
			if ((nState & STATE_SYSTEM_INVISIBLE) == 0)
			{
				if (cchBuf > _tcslen(_T("visible")))
					_tcscpy(lpszBuf, _T("visible"));
			}

			TCHAR szState[200];
			szState[0] = 0;
			UINT nChars = 0;

			// convert state flags to comma separated list
			for (DWORD dwStateBit = STATE_SYSTEM_UNAVAILABLE; 
				 (dwStateBit < STATE_SYSTEM_HASPOPUP) && (nChars < cchBuf); 
				 dwStateBit <<= 1)
			{
				if (nState & dwStateBit)
				{
					szState[0] = 0;
					nChars += GetStateText(dwStateBit, szState, sizeof(szState)/sizeof(TCHAR)-1);
					if ((nChars < (cchBuf-3)) && (szState[0] != 0))
					{
						if (lpszBuf[0] != 0)
							_tcscat(lpszBuf, _T(","));
						_tcscat(lpszBuf, szState);
						nChars = (UINT)_tcslen(lpszBuf);
					}
				}
			}
			bRet = TRUE;
		}
	}

	return bRet;
}

//=============================================================================
BOOL CXMSAALib::EnumAccessible(HWND hwnd,
							   XENUMACCESSIBLEPROC lpEnumAccessibleProc, 
							   LPARAM lParam) 
//=============================================================================
{
	BOOL bRet = FALSE;

	_ASSERTE (::IsWindow(hwnd));
	_ASSERTE(lpEnumAccessibleProc);

	if (::IsWindow(hwnd) && lpEnumAccessibleProc)
	{
		IAccessible *pIAcc = NULL;

		HRESULT hr = AccessibleObjectFromWindow(hwnd, OBJID_WINDOW, 
			IID_IAccessible, (void**)&pIAcc);

		if (SUCCEEDED(hr) && pIAcc)
		{
			VARIANT varChild;
			VariantInit(&varChild);
			IAccessible *pIAccChild = NULL;
			FindChild(pIAcc, &pIAccChild, &varChild, lpEnumAccessibleProc, 
				lParam, TRUE);
			TRACE(_T("return from FindChild\n"));
			SAFE_RELEASE(pIAcc);
			VariantClear(&varChild);
			bRet = TRUE;
		}
	}
	TRACE(_T("exiting EnumAccessible\n"));
	return bRet;
}

//=============================================================================
BOOL CXMSAALib::FindChild(IAccessible *pIAccParent, 
						  IAccessible **pIAccChild, 
						  VARIANT *pvarChild,
						  XENUMACCESSIBLEPROC lpEnumAccessibleProc,
						  LPARAM lParam,
						  BOOL bFirstTime)
//=============================================================================
{
	_ASSERTE(pIAccParent);
	_ASSERTE(pIAccChild);
	_ASSERTE(pvarChild);
	_ASSERTE(lpEnumAccessibleProc);

	IEnumVARIANT *pEnum = NULL;
	VARIANT varChild;
	VariantInit(&varChild);
	IAccessible *pCAcc = NULL;
	BOOL bContinue = TRUE;
	static int nLevel = 0;

	if (bFirstTime)
	{
		nLevel = 0;
		varChild.vt = VT_I4;
		varChild.lVal = CHILDID_SELF;
		bContinue = lpEnumAccessibleProc(pIAccParent, &varChild, NULL, 
						nLevel, lParam);
	}

	nLevel++;

	HRESULT hr = pIAccParent->QueryInterface(IID_IEnumVARIANT, (PVOID*) &pEnum);

	if (pEnum)
		pEnum->Reset();

	// get child count
	long nChildren = 0;
	unsigned long nFetched = 0;

	pIAccParent->get_accChildCount(&nChildren);
	TRACE(_T("nChildren=%d\n"), nChildren);

	for (long index = 1; (index <= nChildren) && bContinue; index++)
	{
		TRACE(_T("index=%d\n"), index);

		VariantClear(&varChild);
		SAFE_RELEASE(pCAcc);

		if (pEnum)
		{
			hr = pEnum->Next(1, &varChild, &nFetched);
			if (!SUCCEEDED(hr))
			{
				bContinue = FALSE;
				break;
			}
		}
		else
		{
			varChild.vt = VT_I4;
			varChild.lVal = index;
		}

		// get IDispatch interface for the child
		IDispatch *pDisp = NULL;
		if (varChild.vt == VT_I4)
		{
			hr = pIAccParent->get_accChild(varChild, &pDisp);
		}
		else if (varChild.vt == VT_DISPATCH)
		{
			pDisp = varChild.pdispVal;
		}

		// get IAccessible interface for the child
		if (pDisp)
		{
			hr = pDisp->QueryInterface(IID_IAccessible, (void**)&pCAcc);
			SAFE_RELEASE(pDisp);
		}

		// get information about the child
		if (pCAcc)
		{
			VariantInit(&varChild);
			varChild.vt = VT_I4;
			varChild.lVal = CHILDID_SELF;
			*pIAccChild = pCAcc;
		}
		else
		{
			*pIAccChild = pIAccParent;
		}

		UINT nState = 0;
		GetObjectState(*pIAccChild, &varChild, nState);

		// check if object is available
		if (nState & (/*STATE_SYSTEM_UNAVAILABLE |*/ STATE_SYSTEM_INVISIBLE /*| STATE_SYSTEM_OFFSCREEN*/))
		{
			SAFE_RELEASE(pCAcc);
			continue;
		}

		UINT nRole = 0;
		GetObjectRole(*pIAccChild, &varChild, nRole);

		HWND hwndChild = 0;
		WindowFromAccessibleObject(*pIAccChild, &hwndChild);

		// call enum callback
		bContinue = lpEnumAccessibleProc(*pIAccChild, &varChild, hwndChild, 
						nLevel, lParam);

		if (bContinue && pCAcc)
		{
			bContinue = FindChild(pCAcc, pIAccChild, pvarChild, 
							lpEnumAccessibleProc, lParam, FALSE);
			if (*pIAccChild != pCAcc)
				SAFE_RELEASE(pCAcc);
		}
	}

	TRACE(_T("============= end of children (%d) ================\n"), nChildren);

	SAFE_RELEASE(pCAcc);
	SAFE_RELEASE(pEnum);
	VariantClear(&varChild);

	nLevel--;

	return bContinue;
}
